/*
 * Base structure
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

/* Move down content because we have a fixed navbar that is 50px tall */
body {
  padding-top: 50px;
}

/*
 * Tags and labels.
 */
.tag {
  margin: 0 3px 3px 0;
}

.tag-default {
  background-color: #efefef !important;
  color: #000 !important;
}

.tag-warning {
  background-color: #f0ad4e !important;
  border-color: #eea236 !important;
  color: #fff !important;
}
.tag-success {
  background-color: #5cb85c !important;
  border-color: #4cae4c !important;
  color: #fff !important;
}

.spark {
  display: inline-block;
}

.spark-legend {
  display: inline-block;
}

.dygraph-legend {
  text-align: right;
}
